
static ERROR ChildFeedback(OBJECTID, struct XMLTag *, STRING, STRING);
static ERROR ConvertXMLTag(struct XMLTag *, STRING, LONG);

/*****************************************************************************
** Function: LoadScript()
*/

static ERROR LoadScript(struct Script *Self)
{
   struct FileInfo info;
   struct acSetUnlistedField set;
   struct XMLTag *tag, **xmllist;
   UBYTE buffer[3000], addargument;
   LONG i, j, k, len, count;
   BYTE dml_processed;
   STRING arg;

   DPrintF("~Script:","[%d] Loading script \"%s\", Proc: %s", Self->Head.UniqueID, Self->Location, Self->Procedure);

   /*** Load the XML file ***/

   if (Self->Location) {
      if (Self->XML) {
         /* Compare date-stamps to see if the file has been altered
         ** since the last time that we used it.
         */

         if (GetFileInfo(Self->Location, &info, sizeof(info)) IS ERR_Okay) {
            if ((info.TimeStamp IS Self->TimeStamp) AND (info.Size IS Self->FileSize)) {
               StepBack();
               return(ERR_Okay);
            }
         }
         else {
            StepBack();
            return(ERR_Okay); /* Do not reload if the original location can no longer be analysed */
         }

         Action(AC_Free, Self->XML, NULL);
         Self->XML = NULL;
      }

      if (CreateObject(ID_XML, NF_CHILD, &Self->XML, NULL,
            FID_Location, Self->Location,
            FID_Flags,    XMF_WELLFORMED|XMF_LOWERCASE,
            NULL) != ERR_Okay) { StepBack(); return(ObjectError(Self, ERH_Function, ERR_CreateObject)); }
   }
   else if (Self->String) {
      if (Self->XML) {
         /*** Do nothing because the string has already been translated to XML ***/
         StepBack();
         return(ERR_Okay);
      }

      if (CreateObject(ID_XML, NF_CHILD, &Self->XML, NULL,
            FID_Statement, Self->String,
            FID_Flags,     XMF_WELLFORMED|XMF_LOWERCASE,
            NULL) != ERR_Okay) { StepBack(); return(ObjectError(Self, ERH_Function, ERR_CreateObject)); }
   }
   else { StepBack(); return(ERR_MissingLocation); }

   /*** Process reserved keywords like <dml> and <args> ***/

   if (GetFields(Self->XML, FID_Tags|TPTR, &xmllist,
                            FID_TimeStamp|TLARGE, &Self->TimeStamp,
                            FID_FileSize|TLARGE,  &Self->FileSize,
                            TAGEND) != ERR_Okay) {
      StepBack();
      return(ObjectError(Self, ERH_Function, ERR_GetField));
   }

   DPrintF("Script:","Processing reserved keywords...");

   dml_processed = FALSE;
   for (count=0; xmllist[count]; count++) {
      tag = xmllist[count];

      if ((StrCompare("dml", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) AND (dml_processed IS FALSE)) {
         /*** Process the <dml> tag ***/

         dml_processed = TRUE;
         for (j=1; j < tag->TotalAttributes; j++) {
            if (StrCompare("type", tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               if (StrCompare("program", tag->Attributes[j].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->Flags |= SCF_PROGRAM;
               }
               else if (StrCompare("batch", tag->Attributes[j].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->Flags |= SCF_BATCH;
               }
               else if (StrCompare("extension", tag->Attributes[j].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->Flags |= SCF_EXTENSION;
               }
               else DPrintF("@Script:","[DML] Unrecognised type '%s'.", buffer);
            }
            else if (StrCompare("service", tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               SetField(Self, FID_Service, FT_STRING, tag->Attributes[j].Value);
            }
         }
      }
      else if (StrCompare("args", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         /*** Process the <args> tag ***/

         GET_LocalArgs(Self, &Self->LocalArgs);

         for (j=1; j < tag->TotalAttributes; j++) {
            addargument = TRUE;
            if (Self->LocalArgs) {
               /* Check if the argument has already been preset.  We don't want
               ** to overwrite argument values that the user has specified.
               */
               for (i=0; (i < Self->TotalArgs) AND (addargument IS TRUE); i++) {
                  if ((arg = AccessArg(Self, i))) {
                     for (len=1, k=0; (arg[len]) AND (arg[len] != '}'); len++, k++) {
                        if (LCase(tag->Attributes[j].Name[k]) != LCase(arg[len])) break;
                     }

                     if ((!tag->Attributes[j].Name[k]) AND (arg[len] IS '}')) addargument = FALSE;

                     ReleaseArg(Self, arg);
                  }
               }
            }

            if (addargument IS TRUE) {
               set.Field = tag->Attributes[j].Name;
               set.Value = tag->Attributes[j].Value;
               SCRIPT_SetUnlistedField(Self, &set);
            }
         }
      }
      else if (StrCompare("translation", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         /* Translation attributes:
         **
         **    Language - The language used in the script is defined here.  If no language
         **               is set, then UK English is assumed.  Language codes are listed in
         **               system:config/languages.cfg
         **    Dir      - The directory of language file storage is defined here.  if not
         **               set, the default is the lang/ directory.
         **
         ** Following the attributes are the translation arguments.
         */

         for (j=1; j < tag->TotalAttributes; j++) {
         	if (StrCompare("language", tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               StrCopy(tag->Attributes[j].Value, Self->Language, sizeof(Self->Language));
            }
            else if (StrCompare("dir", tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               StrCopy(tag->Attributes[j].Value, Self->LanguageDir, sizeof(Self->LanguageDir));
            }
            else {
               set.Field = tag->Attributes[j].Name;
               set.Value = tag->Attributes[j].Value;
               SCRIPT_SetUnlistedField(Self, &set);
            }
         }

         if (StrCompare(glLocale, Self->Language, 0, STR_MATCHLENGTH) != ERR_Okay) {
            /* If the user's language does not match that of the script file,
            ** check for the presence of an acceptable language file in the
            ** script's language folder and use it.
            **
            ** The format is: {scriptdir}/lang/{filename}.languagecode
            **
            ** The translation arguments will be expected to be in the SCRIPT
            ** section of the config file.
            */

            OBJECTPTR config;
            UBYTE filename[300];
            struct ConfigEntry *entries;
            LONG amtentries;

            for (i=0; Self->Location[i]; i++);
            while ((i > 0) AND (Self->Location[i-1] != ':') AND (Self->Location[i-1] != '/') AND (Self->Location[i-1] != '\\')) i--;

            j = CharCopy(Self->Location, filename, i); /* script dir */
            StrFormat(filename+j, sizeof(filename)-j, "%s/%s.%s", Self->LanguageDir, Self->Location + i, glLocale);

            if (AnalyseLocation(filename, NULL) IS ERR_Okay) {
               if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
                  FID_Location|TSTRING, filename,
                  TAGEND) IS ERR_Okay) {

                  if ((GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                                         FID_Entries|TPTR,     &entries,
                                         TAGEND) IS ERR_Okay) AND (entries)) {
                     for (i=0; i < amtentries; i++) {
                        if (StrCompare("SCRIPT", entries[i].Section, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                           set.Field = entries[i].Item;
                           set.Value = entries[i].Data;
                           SCRIPT_SetUnlistedField(Self, &set);
                        }
                     }
                  }

                  Action(AC_Free, config, NULL);
               }
            }
         }
      }
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Function: NextTag()
**
** This is the core routine for processing the individual tag lines.  It is
** based on the reasoning that each line is either going to start with the
** name of a class, a reserved word, or a closing tag.
**
** Currently recognised reserved words:
**
**    <repeat>    <free>     <set>     <xml>
**    <args>      <if>       <dml>     <exit>
**    <unique>    <identify> <break>
**
** If the <set> tag is used without an accompanying object parameter, the
** command will set the specified value(s) against the script object iself.
**
** Currently recognised reserved arguments:
**
**    {index}      {scriptlocation}   {currenttask}
**    {scriptdir}  {id}               {self}
**    {unix}       {owner}/{target}   {scriptowner}
*/

static struct XMLTag * NextTag(struct Script *Self, struct XMLTag *XMLTag, OBJECTID OwnerID,
                               STRING RepeatCount, STRING TagBuffer, STRING Buffer)
{
   struct acDataChannel channel;
   struct mtGetXMLString getxml;
   struct XMLTag *XMLChild;
   OBJECTPTR Object;
   OBJECTID ObjectID;
   LONG count, i, pos, j;
   DOUBLE calc;

   if (!XMLTag) return(NULL);

   if (Self->prvQuit IS TRUE) return(NULL);

   /* If no name is specified then the XML tag is content based.  Content is sent to
   ** parent objects via data channels.
   */

   if (!XMLTag->Attributes[0].Name) {
      /* Do not send the content if it is only whitespace */
      for (i=0; (XMLTag->Attributes[0].Value[i] != 0) AND (XMLTag->Attributes[0].Value[i] <= 0x20); i++);

      if (XMLTag->Attributes[0].Value[i]) {
         if (AccessObject(OwnerID, 3000, &Object) IS ERR_Okay) {
            channel.ObjectID     = Self->Head.UniqueID;
            channel.DataType     = DATA_TEXT;
            channel.Version      = 1;

            if (Self->prvTranslateAll IS TRUE) {
               channel.Buffer = Buffer;
               StrCopy(XMLTag->Attributes[0].Value, Buffer, SIZE_GLOBAL);

               TranslateArgs(Self, Buffer, RepeatCount, SIZE_GLOBAL); /* Translate args */
               StrTranslate(Buffer, SIZE_GLOBAL, STF_IGNOREQUOTES); /* Translate objects */
            }
            else {
               channel.Buffer = XMLTag->Attributes[0].Value;
            }
            channel.Size         = StrLength(channel.Buffer) + 1;
            channel.TotalEntries = 1;
            Action(AC_DataChannel, Object, &channel);
            ReleaseObject(Object);
         }
      }

      return(XMLTag->Next);
   }

   /*** Ignore comments and special definitions identified by ! and ? characters ***/

   if ((*XMLTag->Attributes[0].Name IS '!') OR (*XMLTag->Attributes[0].Name IS '?')) return(XMLTag->Next);

   if (StrCompare("args", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) return(XMLTag->Next);

   /* Check if the owner object exists.  Sometimes during script
   ** processing, the top level owner can be killed if an action in
   ** the script frees it on purpose.  If this occurs, we change the
   ** owner to the SystemDisplay object in order to keep things alive.
   */

   if (OwnerID) {
      if (CheckObjectExists(OwnerID, NULL) != ERR_True) {
         OwnerID = CurrentTaskID();
         DPrintF("Activate:","[DML] The owner was destroyed, so I have switched to task #%d.", OwnerID);
      }
   }
   else {
      OwnerID = CurrentTaskID();
      DPrintF("Activate:","[DML] No owner; using task #%d.", OwnerID);
   }

   /*** Convert the current XML tag into a string ***/

   ConvertXMLTag(XMLTag, TagBuffer, SIZE_GLOBAL);

   /*** Translate arguments in this tag-line ***/

   TranslateArgs(Self, TagBuffer, RepeatCount, SIZE_GLOBAL);

   /*** Translate [owner] and [container] references on the line to the ID equivalent ***/

   for (i=0; TagBuffer[i]; i++) {
      if ((TagBuffer[i] IS '[') AND (TagBuffer[i+1] != '#') AND (TagBuffer[i+1] != '[')) {
         j   = 0;
         pos = i+1; /* <-- Remember the current string position */
         i++;
         while ((TagBuffer[i]) AND (TagBuffer[i] != ']') AND
                (TagBuffer[i] != '[') AND (TagBuffer[i] != '.')) {
            Buffer[j++] = TagBuffer[i++];
         }
         Buffer[j] = 0;

         if ((StrCompare("container", Buffer, NULL, STR_MATCHLENGTH) IS ERR_Okay) OR
             (StrCompare("owner", Buffer, NULL, STR_MATCHLENGTH) IS ERR_Okay)) {
            ObjectID = OwnerID;
            Buffer[0] = '#';
            IntToStr(ObjectID, Buffer+1, SIZE_GLOBAL-1);
            if (StrInsert(Buffer, TagBuffer, pos, j, SIZE_GLOBAL) != ERR_Okay) break;
            i = pos;
         }
      }
   }

   /* Expand any field references that are *outside* of field definitions on
   ** the tag line.  For example:  <script src="blah.dml" [cfgObject.Args(0)]/>
   */

   i = 0;
   while (TagBuffer[i]) {
      if (TagBuffer[i] IS '[') {

         count = 1;
         for (j=0; (TagBuffer[i+j]); j++) {
            Buffer[j] = TagBuffer[i+j];
            if (TagBuffer[i+j] IS '[') count++;
            if (TagBuffer[i+j] IS ']') count--;
            if (!count) break;
         }
         Buffer[j] = 0;

         if (StrTranslate(Buffer, SIZE_GLOBAL, STF_IGNOREQUOTES) IS ERR_Okay) {
            if (StrInsert(Buffer, TagBuffer, i, j, SIZE_GLOBAL) != ERR_Okay) {
               break;
            }

            /*** Translate any arguments resulting from the translation ***/

            for (j=0; (TagBuffer[j]) AND (TagBuffer[j] != '{'); j++);
            if (TagBuffer[j] IS '{') {
               TranslateArgs(Self, TagBuffer, "0", SIZE_GLOBAL);
            }
         }
         else {
            DPrintF("Script:","Translation of \"%s\" failed. (Buffer: %s)", Buffer, TagBuffer);
            return(XMLTag->Next);
         }
      }

      /*** Find the next equal, bracket or space character ***/

      while ((TagBuffer[i]) AND (TagBuffer[i] != '=') AND (TagBuffer[i] != '[')) i++;

      /*** If an equal character was discovered, process the value ***/

      if (TagBuffer[i] IS '=') {
         i++;
         while ((TagBuffer[i]) AND (TagBuffer[i] <= 0x20)) i++; /* skip whitespace */

         if (TagBuffer[i] IS '"') {
            i++;
            while ((TagBuffer[i]) AND (TagBuffer[i] != '"')) i++;
            if (TagBuffer[i]) i++;
         }
         else if (TagBuffer[i] IS '\'') {
            TagBuffer[i] = '"';
            i++;
            while ((TagBuffer[i]) AND (TagBuffer[i] != '\'')) i++;
            if (TagBuffer[i]) { TagBuffer[i] = '"'; i++; }
         }
         else while (TagBuffer[i] > 0x20) i++;

         while ((TagBuffer[i] != 0) AND (TagBuffer[i] <= 0x20)) i++; /* skip whitespace */
      }
   }

   /*** Reserved word processing ***/

   if (StrCompare("xml", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      /* This is the handler for the <xml> tag.  In this instance
      ** we will take all the children in this tag and decompile them
      ** into an XML string which will be sent to the object identified
      ** in the OwnerID.
      **
      ** Please note that if no child tags or content is inside the XML
      ** string, or if attributes are attached to the XML tag, then the
      ** user is trying to create a new XML object (under the Data category),
      ** not the XML reserved word.
      */

      if ((XMLTag->Child) AND (XMLTag->TotalAttributes <= 2)) {
         OBJECTID xml_id;

         if ((i = StrSearch("object=", TagBuffer, FALSE)) != -1) {
            ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);
            xml_id = StrToInt(Buffer);
         }
         else xml_id = OwnerID;

         getxml.Index = XMLTag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         if (Action(MT_GetXMLString, Self->XML, &getxml) IS ERR_Okay) {
            struct MemInfo meminfo;
            if (MemoryPtrInfo(getxml.Result, VER_MemoryInfo, &meminfo) IS ERR_Okay) {
               if (meminfo.Size > SIZE_GLOBAL) {
                  channel.ObjectID     = Self->Head.UniqueID;
                  channel.DataType     = DATA_XML;
                  channel.Version      = 1;
                  channel.Buffer       = getxml.Result;
                  channel.Size         = StrLength(channel.Buffer) + 1;
                  channel.TotalEntries = 1;
                  ActionMsg(AC_DataChannel, xml_id, &channel);
               }
               else {
                  StrCopy(getxml.Result, Buffer, SIZE_GLOBAL);

                  TranslateArgs(Self, Buffer, RepeatCount, SIZE_GLOBAL);

                  channel.ObjectID     = Self->Head.UniqueID;
                  channel.DataType     = DATA_XML;
                  channel.Version      = 1;
                  channel.Buffer       = Buffer;
                  channel.Size         = StrLength(channel.Buffer) + 1;
                  channel.TotalEntries = 1;
                  ActionMsg(AC_DataChannel, xml_id, &channel);
               }
               FreeMemory(getxml.Result);
            }
         }

         return(XMLTag->Next);
      }
   }
   else if (StrCompare("xml-translate", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      OBJECTID xml_id;

      /*** Same as <xml> but translates the content ***/

      if ((i = StrSearch("object=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);
         xml_id = StrToInt(Buffer);
      }
      else xml_id = OwnerID;

      getxml.Index = XMLTag->Child->Index;
      getxml.Flags = XMF_INCLUDESIBLINGS;
      if (Action(MT_GetXMLString, Self->XML, &getxml) IS ERR_Okay) {
         StrCopy(getxml.Result, Buffer, SIZE_GLOBAL);
         FreeMemory(getxml.Result);

         TranslateArgs(Self, Buffer, RepeatCount, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, STF_KEEPINTEGRITY|STF_IGNOREQUOTES);

         channel.DataType     = DATA_XML;
         channel.Version      = 1;
         channel.Buffer       = Buffer;
         channel.Size         = StrLength(channel.Buffer) + 1;
         channel.TotalEntries = 1;
         ActionMsg(AC_DataChannel, xml_id, &channel);
      }
      return(XMLTag->Next);
   }
   else if (StrCompare("break", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      Self->LoopIndex = Self->LoopEnd;
      return(NULL);
   }
   else if (StrCompare("exit", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      Self->prvQuit = TRUE;
      return(NULL);
   }
   else if (StrCompare("dml", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      /*** Skip DML procedures in the script that are not part of the procedure we are executing ***/
      return(XMLTag->Next);
   }
   else if (StrCompare("identify", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      struct XMLTag *xmltag;
      OBJECTPTR class;
      STRING classname;
      LONG classid, subclassid;

      /* The <identify> tag analyses a file location and reports on the class
      ** that is capable of handling the file data.  The developer utilises
      ** this feature in a case-styled statement, as follows:
      **
      **  <identify location="athene:music/mayhem.mp3">
      **    <music>
      **      ...
      **    </music>
      **    <sound>
      **      ...
      **    </sound>
      **  </identify>
      */

      if ((i = StrSearch("location=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);

         if (!Buffer[0]) return(XMLTag->Next);
         if ((Buffer[0] IS '?') AND (Buffer[1] IS 0)) return(XMLTag->Next);

         if (IdentifyFile(Buffer, &classid, &subclassid) IS ERR_Okay) {
            if ((class = FindClass(classid, NULL))) {
               GetFields(class, FID_Name|TSTRING, &classname, TAGEND);

               /* Execute whichever child tag matches the identified class name.  If
               ** no name matches, execute the <other> tag option if the author
               ** has specified one.
               */

               XMLChild = XMLTag->Child;
               xmltag = NULL;
               while (XMLChild) {
                  if (StrCompare(classname, XMLChild->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     xmltag = XMLChild->Child;
                     break;
                  }
                  else if (StrCompare("other", XMLChild->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     xmltag = XMLChild->Child;
                  }
                  XMLChild = XMLChild->Next;
               }

               if (xmltag) {
                  /*** Execute the content of the identified class tag ***/
                  while (xmltag) xmltag = NextTag(Self, xmltag, OwnerID, RepeatCount, TagBuffer, Buffer);
               }
               else DPrintF("@Script:","Identified class \"%s\", .", classname);
            }
            else DPrintF("@Script:","Failed to find class %d.", classid);
         }
         else DPrintF("@Script:","Failed to identify location \"%s\".", Buffer);

         return(XMLTag->Next);
      }
      else return(XMLTag->Next);
   }
   else if (StrCompare("repeat", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      UBYTE rcount[20];
      LONG step, saveindex, saveend;

      saveindex = Self->LoopIndex;
      saveend   = Self->LoopEnd;

      Self->LoopIndex = 0;
      Self->LoopEnd   = 0;
      step  = 1;

      if ((i = StrSearch("start=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);
         StrCalculate(Buffer, &calc, NULL, NULL);
         Self->LoopIndex = calc;
         if (Self->LoopIndex < 0) Self->LoopIndex = 0;
      }

      if ((i = StrSearch("count=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);
         StrCalculate(Buffer, &calc, NULL, NULL);
         if (calc <= 0) calc = 0;
         Self->LoopEnd = Self->LoopIndex + calc;
      }

      if ((i = StrSearch("end=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);
         StrCalculate(Buffer, &calc, NULL, NULL);
         calc = calc + 1;
         if (calc < Self->LoopIndex) Self->LoopEnd = Self->LoopIndex;
         else Self->LoopEnd = calc;
      }

      if ((i = StrSearch("step=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         StrTranslate(Buffer, SIZE_GLOBAL, NULL);
         StrCalculate(Buffer, &calc, NULL, NULL);
         step = calc;
         if (step < 1) step = 1;
      }

      DPrintF("~Activate:","[Script] Performing a repeat loop (start: %d, end: %d, step: %d).", Self->LoopIndex, Self->LoopEnd, step);

      while (Self->LoopIndex < Self->LoopEnd) {
         DPrintF("5Activate:","[Script] Iteration #%d", Self->LoopIndex);
         IntToStr(Self->LoopIndex, rcount, sizeof(rcount)-1);
         XMLChild = XMLTag->Child;
         while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, rcount, TagBuffer, Buffer);
         Self->LoopIndex += step;
      }

      Self->LoopIndex = saveindex;
      Self->LoopEnd   = saveend;

      StepBack();
      DPrintF("Activate:","Repeat loop ends.");

      return(XMLTag->Next);
   }
   else if (StrCompare("free", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if ((i = StrSearch("object=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         if (*Buffer IS '[') {
            i = 1;
            Buffer[StrLength(Buffer)-1] = 0;
         }
         else i = 0;

         if (FastFindObject(Buffer+i, NULL, &ObjectID, 1, &count) IS ERR_Okay) {
            DPrintF("~2Script:","Object #%d will be freed.", ObjectID);
            if (AccessObject(ObjectID, 1000, &Object) IS ERR_Okay) {
               Action(AC_Free, Object, NULL);
               ReleaseObject(Object);
            }
            else ActionMsg(AC_Free, ObjectID, NULL);
            StepBack();
         }
         else DPrintF("@Script:","Error translating line: \"%s\".", TagBuffer);
      }
      else if ((i = StrSearch("content=", TagBuffer, FALSE)) != -1) {
         ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
         if (*Buffer IS '[') {
            i = 1;
            Buffer[StrLength(Buffer)-1] = 0;
         }
         else i = 0;

         if (FastFindObject(Buffer+i, NULL, &ObjectID, 1, &count) IS ERR_Okay) {
            DPrintF("~2Script:","Object #%d's content will be freed.", ObjectID);
            if ((count = TotalChildren(ObjectID))) {
               struct ChildEntry list[count];
               if (ListChildren(ObjectID, list, &count) IS ERR_Okay) {
                  for (i=0; i < count; i++) ActionMsg(AC_Free, list[i].ObjectID, NULL);
               }
            }
            StepBack();
         }
         else DPrintF("@Script:","Error translating line: \"%s\".", TagBuffer);
      }
      else DPrintF("@Script:","Error in line: \"%s\".", TagBuffer);

      if ((XMLChild = XMLTag->Child)) {
         while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, RepeatCount, TagBuffer, Buffer);
      }

      return(XMLTag->Next);
   }
   else if (StrCompare("if", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (StrSearch(" static", TagBuffer, FALSE) IS -1) {
         STRING Test, Compare;
         LONG Condition, satisfied;
         UBYTE save, reverse;
         WORD conditionpos;
         struct File *file;

         satisfied = FALSE;
         reverse = FALSE;

         if (StrSearch(" not ", TagBuffer, FALSE) != -1) reverse = TRUE;

         if ((i = StrSearch(" exists=", TagBuffer, FALSE)) != -1) {
            ExtractValue(TagBuffer+i+1, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);
            if (*Buffer) {
               if (FastFindObject(Buffer, NULL, &ObjectID, 1, &count) IS ERR_Okay) {
                  satisfied = TRUE;
               }
            }
         }
         else if ((i = StrSearch(" fileexists=", TagBuffer, FALSE)) != -1) {\
            LONG flags;

            ExtractValue(TagBuffer+i+1, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);

            if (Buffer[0] IS '~') {
               flags = FL_APPROXIMATE;
               i = 1;
            }
            else {
               flags = NULL;
               i = 0;
            }

            if (CreateObject(ID_FILE, NULL, &file, NULL,
                  FID_Location|TSTRING, Buffer + i,
                  FID_Flags|TLONG,      flags,
                  TAGEND) IS ERR_Okay) {
               satisfied = TRUE;
               Action(AC_Free, file, NULL);
            }
         }
         else if (((i = StrSearch(" directory=", TagBuffer, FALSE)) != -1) OR
                  ((i = StrSearch(" isdirectory=", TagBuffer, FALSE)) != -1)) {
            /*** This option checks if a location explicitly refers to a directory ***/
            ExtractValue(TagBuffer+i+1, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);
            if (CreateObject(ID_FILE, NULL, &file, NULL,
                  FID_Location|TSTRING, Buffer,
                  TAGEND) IS ERR_Okay) {
               if (file->Flags & FL_DIRECTORY) satisfied = TRUE;
               Action(AC_Free, file, NULL);
            }
         }
         else if ((i = StrSearch(" isnull=", TagBuffer, FALSE)) != -1) {
            ExtractValue(TagBuffer+i+1, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);
            if ((!Buffer[0]) OR (*Buffer IS '0')) satisfied = TRUE;
            else satisfied = FALSE;
         }
         else if ((i = StrSearch(" notnull=", TagBuffer, FALSE)) != -1) {
            ExtractValue(TagBuffer+i+1, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);
            if ((!Buffer[0]) OR (*Buffer IS '0')) satisfied = FALSE;
            else satisfied = TRUE;
         }
         else {
            Condition = NULL;
            Compare   = NULL;
            Test      = NULL;

            /*** Statement field ***/

            if ((i = StrSearch("statement=", TagBuffer, FALSE)) != -1) {
               ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
               StrTranslate(Buffer, SIZE_GLOBAL, NULL);

               /* If the statement starts with a calculation (=), do
               ** not confuse it with a condition character.
               */

               if (Buffer[0] IS '=') i = 1;
               else i = 0;

               /*** Find the condition statement ***/

               while (Buffer[i]) {
                  if ((Buffer[i] IS '!') AND (Buffer[i+1] IS '=')) break;
                  if (Buffer[i] IS '>') break;
                  if (Buffer[i] IS '<') break;
                  if (Buffer[i] IS '=') break;
                  i++;
               }

               /*** Continue if a condition was found ***/

               if (Buffer[i]) {
                  conditionpos = i;

                  /*** Test field ***/

                  while ((i > 0) AND (Buffer[i-1] IS ' ')) i--;
                  save = Buffer[i];
                  Buffer[i] = 0;
                  if (*Buffer IS '=') {
                     UBYTE tmp[1024];
                     StrCalculate(Buffer+1, NULL, tmp, sizeof(tmp));
                     Test = StrClone(tmp);
                  }
                  else Test = StrClone(Buffer);
                  Buffer[i] = save;

                  /*** Condition field ***/

                  for (i=conditionpos; (Buffer[i] IS '!') OR (Buffer[i] IS '=') OR (Buffer[i] IS '>') OR (Buffer[i] IS '<'); i++);

                  save = Buffer[i];
                  Buffer[i] = 0;
                  for (j=0; ConditionTable[j].Name; j++) {
                     if (StrCompare(Buffer+conditionpos, ConditionTable[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        Condition = ConditionTable[j].Value;
                        break;
                     }
                  }
                  Buffer[i] = save;

                  /*** Compare field ***/

                  while ((Buffer[i]) AND (Buffer[i] <= 0x20)) i++; /* skip whitespace */

                  if (Buffer[i]) Compare = StrClone(Buffer+i);

                  if (Test) {
                     if (Compare) {
                        if (Condition) {
                           satisfied = TestStatement(Test, Compare, Condition);
                        }
                        else DPrintF("@If:","Missing condition definition in statement \"%s\".", Buffer);
                     }
                     else DPrintF("@If:","Missing compare definition in statement \"%s\".", Buffer);
                  }
                  else DPrintF("@If:","Missing test definition in statement \"%s\".", Buffer);

                  if (Test)    FreeMemory(Test);
                  if (Compare) FreeMemory(Compare);
               }
            }
            else {
               /*** Condition field ***/
               if ((i = StrSearch("condition=", TagBuffer, FALSE)) != -1) {
                  ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
                  for (i=0; ConditionTable[i].Name != NULL; i++) {
                     if (StrCompare(Buffer, ConditionTable[i].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                        Condition = ConditionTable[i].Value;
                        break;
                     }
                  }
               }

               /*** Test field ***/
               if ((i = StrSearch("test=", TagBuffer, FALSE)) != -1) {
                  ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
                  StrTranslate(Buffer, SIZE_GLOBAL, NULL);
                  if (*Buffer IS '=') {
                     UBYTE tmp[1024];
                     StrCalculate(Buffer+1, NULL, tmp, sizeof(tmp));
                     Test = StrClone(tmp);
                  }
                  else Test = StrClone(Buffer);
               }

               /*** Compare field ***/
               if ((i = StrSearch("compare=", TagBuffer, FALSE)) != -1) {
                  ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
                  StrTranslate(Buffer, SIZE_GLOBAL, NULL);
                  if (*Buffer IS '=') {
                     UBYTE tmp[1024];
                     StrCalculate(Buffer+1, NULL, tmp, sizeof(tmp));
                     Compare = StrClone(tmp);
                  }
                  else Compare = StrClone(Buffer);
               }

               if ((Test) AND (Compare) AND (Condition)) {
                  satisfied = TestStatement(Test, Compare, Condition);
               }
               else DPrintF("@If:","Missing test, compare or condition definitions.");

               if (Test)    FreeMemory(Test);
               if (Compare) FreeMemory(Compare);
            }
         }

         /*** If the 'not' option was used, reverse the satisfied flag ***/

         if (reverse) satisfied ^= 1;

         /* React according to the satisfied value.  We will either process everything
         ** inside the if statement, or we will avoid it.  If an else statement follows,
         ** we will process it if the if statement was not satisfied.
         */

         if (satisfied IS TRUE) {
            XMLChild = XMLTag->Child;
            while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, RepeatCount, TagBuffer, Buffer);

            XMLTag = XMLTag->Next;
            if ((XMLTag) AND (StrCompare("else", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay)) XMLTag = XMLTag->Next;
            return(XMLTag);
         }
         else {
            XMLTag = XMLTag->Next;
            if ((XMLTag) AND (StrCompare("else", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
               XMLChild = XMLTag->Child;
               while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, RepeatCount, TagBuffer, Buffer);
               return(XMLTag->Next);
            }
            else return(XMLTag);
         }
      } /* else: continue on to create a -static- if object */
   }
   else if (StrCompare("unique", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      Self->prvUnique++;
      if ((XMLChild = XMLTag->Child)) {
         while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, RepeatCount, TagBuffer, Buffer);
      }
      Self->prvUnique--;
      return(XMLTag->Next);
   }
   else if (StrCompare("set", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      /* NB: If any part of this routine fails in the early stages, we
      ** simply break in order to have the real Set class process the
      ** request instead of this faster routine.
      */

      if (StrSearch(" static", TagBuffer, FALSE) IS -1) {
         OBJECTPTR Object;
         BYTE FieldName[40];

         if ((i = StrSearch("object=", TagBuffer, FALSE)) != -1) {

            ObjectID = NULL;

            ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);
            StrTranslate(Buffer, SIZE_GLOBAL, NULL);

            if (*Buffer) {
               if (StrCompare("owner", Buffer, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  ObjectID = OwnerID;
               }
               else if (StrCompare("container", Buffer, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  ObjectID = OwnerID;
               }
               else if (FastFindObject(Buffer, NULL, &ObjectID, 1, &count) != ERR_Okay) {
                  DPrintF("@DML:","Failed to extract object from: %s", TagBuffer);
               }
            }
         }
         else ObjectID = Self->Head.UniqueID;

         if (ObjectID) {
            if (AccessObject(ObjectID, 5000, &Object) IS ERR_Okay) {
               i = 4;
               while (TagBuffer[i]) {
                  i += ExtractField(FieldName, TagBuffer+i);
                  if (StrCompare("object", FieldName, NULL, STR_MATCHLENGTH) != ERR_Okay) {
                     ExtractValue(TagBuffer+i, Buffer, SIZE_GLOBAL);

                     if (FieldName[0] IS '&') WriteField(Self, Object, FieldName+1, Buffer);
                     else WriteField(Self, Object, FieldName, Buffer);
                  }
                  i += NextElement(TagBuffer+i);
               }

               ReleaseObject(Object);

               if ((XMLChild = XMLTag->Child)) {
                  while (XMLChild) XMLChild = NextTag(Self, XMLChild, OwnerID, RepeatCount, TagBuffer, Buffer);
               }
               return(XMLTag->Next);
            }
            else DPrintF("@Script:","Could not gain access to object %d.", ObjectID);
         }
      }
   }

   /*** Build the object based on the tag information ***/

   if (*TagBuffer IS '*') {
      ObjectID = ProcessObject(Self, OwnerID, TagBuffer+1, Buffer, FALSE);
      Self->PrimaryObjectID = ObjectID;
   }
   else if (*TagBuffer IS '@') {
      ObjectID = ProcessObject(Self, OwnerID, TagBuffer+1, Buffer, TRUE);
   }
   else ObjectID = ProcessObject(Self, OwnerID, TagBuffer, Buffer, FALSE);

   /*** Skip to the next tag if no object was created ***/

   if (!ObjectID) return(XMLTag->Next);

   /*** Initialise the object ***/

   if (AccessObject(ObjectID, 3000, &Object) IS ERR_Okay) {

      if ((Self->Flags & SCF_OBJECTLIST) AND (Object->ClassID IS ID_SCRIPT)) {
         ((struct Script *)Object)->Flags |= SCF_OBJECTLIST;
      }

      if (acInit(Object) IS ERR_Okay) {

         /* If the object is a script and we have children that have to be
         ** initialised to it, set the ChildFeedback() function in the
         ** script object so that it will know to call us when it encounters
         ** the child entry point (marked by an asterisk).
         */

         if ((Object->ClassID IS ID_SCRIPT) AND (XMLTag->Child)) {
            glPrevScript  = glScript;
            glScript      = Self;
            glChildTags   = XMLTag->Child;
            glRepeatCount = RepeatCount;
            ((struct Script *)Object)->ChildFeedback = &ChildFeedback;
            acOpeningTag(Object);
            ((struct Script *)Object)->ChildFeedback = NULL;

            if (Self->Flags & SCF_OBJECTLIST) SetField(Self, FID_RegisterObject, FT_LONG, &ObjectID);
         }
         else {
            /*** Send the OpeningTag signal to the object ***/

            acOpeningTag(Object);
            if (Self->Flags & SCF_OBJECTLIST) SetField(Self, FID_RegisterObject, FT_LONG, &ObjectID);
         }

         /* If we are making an objectlist and the created object is a script or
         ** template, pull the first object or the primary object from that
         ** script to add to our collection.
         */

         if ((Self->Flags & SCF_OBJECTLIST) AND (Object->ClassID IS ID_SCRIPT)) {
            OBJECTID *list;
            if (((struct Script *)Object)->PrimaryObjectID) {
               SetField(Self, FID_RegisterObject, FT_LONG, &((struct Script *)Object)->PrimaryObjectID);
            }
            else if (GetField(Object, FID_ObjectList, FT_POINTER, &list) IS ERR_Okay) {
               SetField(Self, FID_RegisterObject, FT_LONG, list);
            }
         }

         /*** If there is a QUIT message on the message queue, terminate the script process ***/

         if (GetMessage(NULL, MSGID_QUIT, MF_QUERY, NULL, NULL) IS ERR_Okay) {
            DPrintF("Script:","Quit message detected in the local queue, stopping script process.");
            Self->prvQuit = TRUE;
            ReleaseObject(Object);
            return(NULL);
         }

         /* Process child tags.  Note that if the object is a script then the
         ** children would have already been initialised through the child
         ** feedback mechanism.
         */

         if ((XMLChild = XMLTag->Child) AND (GetClassID(ObjectID) != ID_SCRIPT)) {
            while (XMLChild) {
               XMLChild = NextTag(Self, XMLChild, ObjectID, RepeatCount, TagBuffer, Buffer);
            }
         }

         /* If child feedback is required and this is a primary object,
         ** call the ChildFeedback() routine of our parent.
         */

         if ((Self->PrimaryObjectID IS ObjectID) AND (Self->ChildFeedback)) {
            Self->ChildFeedback(Self->PrimaryObjectID, XMLTag->Child, TagBuffer, Buffer);
            Self->ChildFeedback = NULL;
         }

         /*** Send the object the closing tag ***/

         Action(AC_ClosingTag, Object, NULL);
      }
      else {
         /* In the event of failure, we free the object and skip any
         ** children that were intended to be initalised to it before
         ** continuing on with the script.
         */

         DPrintF("@DML:","Object %d failed, it will be freed and skipped.", ObjectID);

         if ((Self->Flags & SCF_EXITONERROR) OR (Self->PrimaryObjectID IS ObjectID)) {
            Self->Error   = ERR_Init;
            Self->prvQuit = TRUE;
         }

         Action(AC_Free, Object, NULL);
         ReleaseObject(Object);
         return(XMLTag->Next);
      }

      ReleaseObject(Object);
   }
   else {
      DPrintF("@Script:","Could not gain access to object #%d.", ObjectID);
      return(NULL);
   }

   /*** If there is a QUIT message on the message queue, terminate the script process ***/

   if (GetMessage(NULL, MSGID_QUIT, MF_QUERY, NULL, NULL) IS ERR_Okay) {
      DPrintF("Script:","Quit message detected in the local queue, stopping script process.");
      Self->prvQuit = TRUE;
      return(NULL);
   }

   return(XMLTag->Next);
}

/*****************************************************************************
** Internal: ChildFeedback()
**
** This function is called when a script is ready to have child objects
** initialised to it.
*/

static ERROR ChildFeedback(OBJECTID OwnerID, struct XMLTag *XMLTag, STRING TagBuffer, STRING Buffer)
{
   struct XMLTag *tags;

   tags = glChildTags;
   while (tags) {
      tags = NextTag(glScript, tags, OwnerID, glRepeatCount, TagBuffer, Buffer);
   }
   glScript = glPrevScript;
   return(ERR_Okay);
}

/*****************************************************************************
** Function: ExtractField()
*/

static LONG ExtractField(STRING FieldName, STRING Source)
{
   WORD i, j;

   for (i=0; (Source[i]) AND (Source[i] <= 0x20); i++);
   for (j=0; (j < 100) AND (Source[i] != '=') AND (Source[i] > 0x20); j++) FieldName[j] = Source[i++];
   FieldName[j] = 0;
   return(i);
}

/*****************************************************************************
** Function: ExtractValue()
**
** This function can be sent strings in this -exact- format:
**
**   myvalue="blah blah blah"
**
** If a value such as the following is sent with no = sign after the text:
**
**   myvalue
**
**  Then the value will be set to a default of TRUE (1).
*/

static void ExtractValue(STRING str, STRING Value, LONG BufferLen)
{
   LONG i, j;

   BufferLen--; /* Subtract 1 for the null-termination byte */

   Value[0] = '1';
   Value[1] = 0;

   while ((*str > 0x20) AND (*str != '=') AND (*str)) str++;

   if (*str IS '=') {
      str += 2; /* Skip the equal sign and the starting quote */
      for (i=0; (*str) AND (*str != '"'); i++) {
         Value[i] = *str++;
         if (i >= BufferLen) { DPrintF("@DML:","Buffer overflow."); break; }
      }
      Value[i] = 0;
   }

   /*** Convert use of "\n" to an actual line feed ***/

   j = 0;
   for (i=0; Value[i]; i++) {
      if ((Value[i] IS '\\') AND (Value[i+1] IS '\\')) {
         Value[j++] = '\\';
         i++;
      }
      else if ((Value[i] IS '\\') AND (Value[i+1] IS 'n')) {
         Value[j++] = '\n';
         i++;
      }
      else Value[j++] = Value[i];
   }
   Value[j] = 0;
}

/*****************************************************************************
** Function: WriteField()
*/

static void WriteField(struct Script *Self, OBJECTPTR Object, STRING FieldName, STRING Buffer)
{
   OBJECTPTR Context;
   LONG len, i;
   UBYTE tmp[1024];

   if ((!Object) OR (!FieldName) OR (!Buffer)) return;

   Context = SetContext(Object);

      /* If the value is encapsulated with dollar symbols '$', or the
      ** field name contains a $ character, then we do not translate
      ** it (we just remove the dollar symbols).
      */

      for (len=0; Buffer[len]; len++);
      if (((Buffer[0] IS '$') AND (Buffer[len-1] IS '$')) OR
          (FieldName[0] IS '$') OR (FieldName[1] IS '$')) {
         Buffer[len-1] = 0;
         SetDMLField(Object, FieldName, Buffer+1);
      }
      else if (StrTranslate(Buffer, SIZE_GLOBAL, NULL) IS ERR_Okay) {

         /* Late argument translation - important if StrTranslate()
         ** reveals new arguments.
         */
         for (i=0; (Buffer[i]) AND (Buffer[i] != '{'); i++);
         if (Buffer[i] IS '{') {
            TranslateArgs(Self, Buffer, "0", SIZE_GLOBAL);
         }

         for (len=0; Buffer[len]; len++);
         if (Buffer[0] IS '=') {
            if (StrCalculate(Buffer+1, NULL, tmp, sizeof(tmp)) IS ERR_Okay) {
               SetDMLField(Object, FieldName, tmp);
            }
         }
         else SetDMLField(Object, FieldName, Buffer);
      }
      else if (Buffer[0] IS '=') {
         if (StrCalculate(Buffer+1, NULL, tmp, sizeof(tmp)) IS ERR_Okay) {
            SetDMLField(Object, FieldName, tmp);
         }
      }
      else SetDMLField(Object, FieldName, Buffer);

   SetContext(Context);
}

/*****************************************************************************
** Internal: SetDMLField()
*/

static void SetDMLField(OBJECTPTR Object, STRING FieldName, STRING Buffer)
{
   struct acSetUnlistedField unlisted;
   FIELD FieldID;
   LONG i;

   /*** If the field value is a single question mark, do nothing ***/

   if ((Buffer[0] IS '?') AND (Buffer[1] IS 0)) return;

   /* Two question marks may be used if a field value wants to legitimately
   ** specify a single question mark in a string.
   */

   if ((Buffer[0] IS '?') AND (Buffer[1] IS '?') AND (Buffer[2] IS 0)) Buffer = "?";

   /*** If the field name is preceded by an ampersand, use unlisted fields ***/

   if (FieldName[0] IS '&') {
      while (*FieldName IS '&') FieldName++;
      unlisted.Field = FieldName;
      unlisted.Value = Buffer;
      Action(AC_SetUnlistedField, Object, &unlisted);
   }
   else {
      /* Check if the field name contains a dot, in which case it
      ** refers to the field of a child object.
      */

      for (i=0; FieldName[i]; i++) {
         if (FieldName[i] IS '.') {
            FieldName[i] = 0;
            if (ResolveFields(FieldName, &FieldID, TAGEND) != ERR_Okay) { FieldName[i] = '.'; return; }
            if ((GetField(Object, FieldID, FT_POINTER, &Object) != ERR_Okay) OR (!Object)) { FieldName[i] = '.'; return; }
            FieldName[i] = '.';
            FieldName += i + 1;
            i = -1;
         }
      }

      /*** Set the field ***/
      SetFieldVariable(Object, FieldName, Buffer);
   }
}

/*****************************************************************************
** Function: NextElement()
*/

static LONG NextElement(STRING str)
{
   LONG i;

   i = 0;
   while ((str[i] <= 0x20) AND (str[i])) i++;
   if (str[i] IS '=') {
      i += 2;
      while ((str[i]) AND (str[i] != '"')) i++;
      if (str[i]) i++;
   }
   return(i);
}

/****************************************************************************/

static LONG TestStatement(STRING TestString, STRING CompareString, LONG Condition)
{
   LONG CompareType, TestType;
   DOUBLE CompareFloat, TestFloat;
   LONG Satisfied;

   //DPrintF("TestStatement()","\"%s\" %d \"%s\"", TestString, Condition, CompareString);

   /*** Convert the If->Compare to its specified type ***/

   CompareType = StrDatatype(CompareString);
   TestType    = StrDatatype(TestString);

   /*** Do the comparison ***/

   Satisfied  = FALSE;
   if (((TestType IS STR_NUMBER) OR (TestType IS STR_FLOAT)) AND ((CompareType IS STR_NUMBER) OR (CompareType IS STR_FLOAT))) {
      CompareFloat = StrToFloat(CompareString);
      TestFloat    = StrToFloat(TestString);
      switch(Condition) {
         case COND_NOT_EQUAL:     if (TestFloat != CompareFloat) Satisfied = TRUE; break;
         case COND_EQUAL:         if (TestFloat IS CompareFloat) Satisfied = TRUE; break;
         case COND_LESS_THAN:     if (TestFloat <  CompareFloat) Satisfied = TRUE; break;
         case COND_LESS_EQUAL:    if (TestFloat <= CompareFloat) Satisfied = TRUE; break;
         case COND_GREATER_THAN:  if (TestFloat >  CompareFloat) Satisfied = TRUE; break;
         case COND_GREATER_EQUAL: if (TestFloat >= CompareFloat) Satisfied = TRUE; break;
         default: DPrintF("@If:","Unsupported condition type %d.", Condition);
      }
   }
   else {
      if (Condition IS COND_EQUAL) {
         if (StrCompare(TestString, CompareString, NULL, STR_MATCHLENGTH) IS ERR_Okay) Satisfied = TRUE;
      }
      else if (Condition IS COND_NOT_EQUAL) {
         if (StrCompare(TestString, CompareString, NULL, STR_MATCHLENGTH) != ERR_Okay) Satisfied = TRUE;
      }
      else DPrintF("@If:","String comparison for condition %d not possible.", Condition);
   }

   return(Satisfied);
}

/****************************************************************************
** Function: ProcessObject()
** Synopsis: OBJECTID ProcessObject(*Script, OBJECTID OwnerID, STRING Tag, STRING Buffer)
*/

static OBJECTID ProcessObject(struct Script *Self, OBJECTID OwnerID, STRING Tag, STRING Buffer, WORD Public)
{
   BYTE fieldname[60];
   STRING name, search;
   LONG ClassID, i, flags, filetype;
   ERROR error;
   UBYTE template;
   OBJECTID ObjectID;
   OBJECTPTR object, owner;

   if ((!OwnerID) OR (!Tag)) return(NULL);

   /*** Extract the class name from the tag ***/

   for (i=0; (*Tag > 0x20) AND (i < 80); i++) Buffer[i] = *Tag++;
   Buffer[i] = 0;

   if (!(ClassID = ResolveClassName(Buffer))) {
      ClassID  = ID_SCRIPT;
      flags    = NF_PRIVATE;
      template = TRUE;

      for (i=0; Buffer[i] AND (Buffer[i] != ':'); i++);

      if (Buffer[i] != ':') StrInsert("templates:", Buffer, 0, 0, SIZE_GLOBAL);

      for (i=0; Buffer[i]; i++);
      if (StrCompare(".dml", Buffer+i-4, NULL, STR_MATCHLENGTH) != ERR_Okay) {
         StrCopy(".dml", Buffer+i, COPY_ALL);
      }

      /*** Check that the script location exists ***/

      if ((AnalyseLocation(Buffer, &filetype) != ERR_Okay) OR (filetype != LOC_FILE)) {
         if (Self->Flags & SCF_EXITONERROR) {
            Self->Error = ObjectError(Self, ERH_Activate, ERR_AccessObject);
            Self->prvQuit = TRUE;
         }
         return(NULL);
      }
   }
   else {
      /* Public objects can be accessed and manipulated by any task that is within the
      ** context of the process that created said public objects.
      */
      if (Public IS TRUE) flags = NF_PUBLIC;
      else flags = NF_PRIVATE;
      template = FALSE;
   }

   /* If the unique option has been specified, we need to extract the name of the object
   ** from the field list for the purposes of creating the unique object.
   */

   name = NULL;
   if ((Self->prvUnique > 0) AND (template IS FALSE)) {
      flags |= NF_UNIQUE;
      search = Tag;
      while (*search) {
         skipwhitespace(search);
         if (!search[0]) break;
         search += ExtractField(fieldname, search);
         ExtractValue(search, Buffer, SIZE_GLOBAL);
         if (StrCompare("name", fieldname, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            name = Buffer;
            break;
         }
         search += NextElement(search);
      }
   }

   /*** Create the new object ***/

   ObjectID = NULL;
   if ((error = NewNamedObject(ClassID, flags, &object, &ObjectID, name)) IS ERR_Okay) {
      if (template IS TRUE) SetField(object, FID_Location, FT_STRING, Buffer);

      /* Change the context to that of the owner as we do not want
      ** our new object to be a child of our Script object.
      */

      if (AccessObject(OwnerID, 3000, &owner) IS ERR_Okay) {
         SetOwner(object, owner);
         ReleaseObject(owner);

         /* Process the fields that have been set against this tag/object
         ** before returning for initialisation.
         */

         while (*Tag) {
            skipwhitespace(Tag);

            if (*Tag IS NULL) break;

            if (template IS TRUE) {
               /* Force use of unlisted fields for template/scripts to
               ** prevent development confusion.
               */
               *fieldname = '&';
               Tag += ExtractField(fieldname+1, Tag);
               ExtractValue(Tag, Buffer, SIZE_GLOBAL);

               if ((StrCompare("static", fieldname+1, NULL, STR_MATCHLENGTH) IS ERR_Okay) OR
                   (StrCompare("target", fieldname+1, NULL, STR_MATCHLENGTH) IS ERR_Okay)) {
                  WriteField(Self, object, fieldname+1, Buffer);
               }
               else WriteField(Self, object, fieldname, Buffer);
            }
            else {
               Tag += ExtractField(fieldname, Tag);
               ExtractValue(Tag, Buffer, SIZE_GLOBAL);
               WriteField(Self, object, fieldname, Buffer);
            }

            Tag += NextElement(Tag);
         }
      }
      else {
         Action(AC_Free, object, NULL);
         if (Self->Flags & SCF_EXITONERROR) {
            Self->Error = ObjectError(Self, ERH_Activate, ERR_AccessObject);
            Self->prvQuit = TRUE;
         }
      }
      ReleaseObject(object);
   }
   else {
      if (error != ERR_ObjectExists) {
         if (Self->Flags & SCF_EXITONERROR) {
            Self->prvQuit = TRUE;
            Self->Error = ObjectError(Self, ERH_Activate, ERR_NewObject);
         }
      }
      ObjectID = NULL;
   }

   return(ObjectID);
}

/*****************************************************************************
** Internal: TranslateArgs()
**
** Translates arguments within the available buffer space of the string.
*/

static void TranslateArgs(struct Script *Self, STRING Buffer, STRING RepeatCount, LONG BufferSize)
{
   STRING arg;
   LONG pos, len, index, count, i, j, k;
   UBYTE str[30], save;

   /* Search for an argument.  If there are no arguments
   ** in the string, return immediately.
   */

   for (pos=0; (Buffer[pos]) AND (Buffer[pos] != '{'); pos++);
   if (Buffer[pos] != '{') return;

   GET_LocalArgs(Self, &Self->LocalArgs);

   count = 0;
   for (i=pos; (Buffer[i]); i++) {
      if (Buffer[i] IS '{') {
         /*** Check against user arguments ***/

         for (index=0; index < Self->TotalArgs; index++) {
            if ((arg = AccessArg(Self, index))) {
               if (StrCompare(arg, Buffer+i, 0, NULL) IS ERR_Okay) {
                  for (len=0; arg[len] != 0; len++);
                  count = 0;

                  StrInsert(arg + len + 1, Buffer, i, len, BufferSize);

                  ReleaseArg(Self, arg);
                  break;
               }
               else ReleaseArg(Self, arg);
            }
         }

         if (Buffer[i] != '{') continue;

         /*** Check against special arguments ***/

         if (StrCompare("{index}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /* Translate {index} arguments to the correct number if
            ** we are inside a repeat loop.  Otherwise, do not translate (as
            ** the {index} could refer to an actual argument).
            */

            //if (RepeatCount) {
               StrInsert(RepeatCount, Buffer, i, sizeof("{index}")-1, BufferSize);
            //}
         }
         else if (StrCompare("{id}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {id} arguments to the correct number ***/
            IntToStr(Self->GeneratedID, str, sizeof(str));
            StrInsert(str, Buffer, i, sizeof("{id}")-1, BufferSize);
         }
         else if (StrCompare("{self}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {self} arguments to the script object's ID ***/
            str[0] = '#';
            IntToStr(Self->Head.UniqueID, str+1, sizeof(str)-1);
            StrInsert(str, Buffer, i, sizeof("{self}")-1, BufferSize);
         }
         else if (StrCompare("{target}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {target} arguments to the target that was used for the script ***/
            if (Self->TargetID IS CurrentTaskID()) {
               StrInsert("0", Buffer, i, sizeof("{target}")-1, BufferSize);
            }
            else {
               IntToStr(Self->TargetID, str, sizeof(str));
               StrInsert(str, Buffer, i, sizeof("{target}")-1, BufferSize);
            }
         }
         else if (StrCompare("{currenttask}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {currenttask} arguments to the correct string ***/
            str[0] = '#';
            IntToStr(CurrentTaskID(), str+1, sizeof(str)-1);
            StrInsert(str, Buffer, i, sizeof("{currenttask}")-1, BufferSize);
         }
         else if (StrCompare("{unix}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Return '1' if the host is a Unix flavoured system ***/
            #if __linux__ || __unix__ ||__bsd__
               StrInsert("1", Buffer, i, sizeof("{unix}")-1, BufferSize);
            #else
               StrInsert("0", Buffer, i, sizeof("{unix}")-1, BufferSize);
            #endif
         }
         else if (StrCompare("{scriptowner}", Buffer+i, 0, NULL) IS ERR_Okay) {
            str[0] = '#';
            IntToStr(Self->ScriptOwnerID, str+1, sizeof(str)-1);
            StrInsert(str, Buffer, i, sizeof("{scriptowner}")-1, BufferSize);
         }
         else if (StrCompare("{platform}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Return the name of the host platform ***/
            #ifdef __linux__
               StrInsert("Linux", Buffer, i, sizeof("{platform}")-1, BufferSize);
            #elif _WIN32
               StrInsert("Windows", Buffer, i, sizeof("{platform}")-1, BufferSize);
            #elif __bsd__
               StrInsert("BSD", Buffer, i, sizeof("{platform}")-1, BufferSize);
            #else
               StrInsert("Athenyx", Buffer, i, sizeof("{platform}")-1, BufferSize);
            #endif
         }
         else if (StrCompare("{scriptlocation}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {scriptlocation} arguments to the correct string ***/
            StrInsert(Self->Location, Buffer, i, sizeof("{scriptlocation}")-1, BufferSize);
         }
         else if (StrCompare("{scriptdir}", Buffer+i, 0, NULL) IS ERR_Okay) {
            /*** Translate {scriptdir} arguments to the correct string ***/
            if ((!Self->Location) AND (Self->LocationMID)) {
               AccessMemory(Self->LocationMID, MEM_READWRITE, (void **)&Self->Location);
            }

            if (Self->Location) {
               j = 0;
               for (k=0; Self->Location[k]; k++) {
                  if ((Self->Location[k] IS ':') OR (Self->Location[k] IS '/') OR (Self->Location[k] IS '\\')) j = k+1;
               }

               save = Self->Location[j];
               Self->Location[j] = 0;
               StrInsert(Self->Location, Buffer, i, sizeof("{scriptdir}")-1, BufferSize);
               Self->Location[j] = save;
            }
            else DPrintF("@DML:","Detected {scriptdir}, but the script object has no Location.");
         }

         if (++count > 60) {
            DPrintF("@TranslateArgs:","[Script] Recursion in line %.20s", Buffer);
            break;
         }
      }
   }
}

/*****************************************************************************
** Internal: ConvertXMLTag()
** Short:    Converts an XMLTag structure into a string.
*/

static ERROR ConvertXMLTag(struct XMLTag *XMLTag, STRING Buffer, LONG BufferSize)
{
   LONG i, j, pos;

   BufferSize--; /* Compensate for null terminator */

   pos = 0;
   for (i=0; i < XMLTag->TotalAttributes; i++) {
      for (j=0; (XMLTag->Attributes[i].Name[j]) AND (pos < BufferSize); j++) Buffer[pos++] = XMLTag->Attributes[i].Name[j];

      if (XMLTag->Attributes[i].Value) {
         if (pos < BufferSize) Buffer[pos++] = '=';
         if (pos < BufferSize) Buffer[pos++] = '"';
         for (j=0; (XMLTag->Attributes[i].Value[j]) AND (pos < BufferSize); j++) {
            if (XMLTag->Attributes[i].Value[j] IS '"') Buffer[pos++] = '\'';
            else Buffer[pos++] = XMLTag->Attributes[i].Value[j];
         }
         if (pos < BufferSize) Buffer[pos++] = '"';
      }

      if ((i+1 < XMLTag->TotalAttributes) AND (pos < BufferSize)) Buffer[pos++] = ' ';
   }

   Buffer[pos] = 0;
   return(ERR_Okay);
}
