#ifndef DATA_XML_H
#define DATA_XML_H TRUE

/*
**  xml.h
**
**  (C) Copyright 2001-2004 Rocklyte Systems.
**      All Rights Reserved.
*/

#ifndef MAIN_H
#include <pandora/main.h>
#endif

/*****************************************************************************
** XML class.
*/

#define VER_XML 1.0

struct XMLAttribute {
   STRING Name;
   STRING Value;
};

struct XMLTag {
   LONG   Index;                    /* Position within the XML array */
   struct XMLTag *Child;            /* Reference to further child tags */
   struct XMLTag *Prev;             /* Reference to the previous tag at this level in the chain */
   struct XMLTag *Next;             /* Reference to the next tag at this level in the chain */
   struct XMLAttribute *Attributes; /* Attributes of the tag, starting with the name */
   WORD   TotalAttributes;          /* Total number of listed attributes for this tag */
   WORD   _Unused;                  /* */
   APTR   Private;                  /* Developer's private memory reference */

   PRIVATE_FIELDS
   LONG   AttribSize;               /* The length of all attribute strings, compressed together */
};

/*****************************************************************************
** Flags.
*/

#define XMF_WELLFORMED      0x00000001 /* Only process an XML statement if it is well formed */
#define XMF_INCLUDECOMMENTS 0x00000002 /* Include comments <!-- --> in the XML tag list */
#define XMF_STRIPCONTENT    0x00000004 /* Do not include the content in amongst the XML tags */
#define XMF_LOWERCASE       0x00000008 /* Force all names and attributes into lower-case characters */
#define XMF_UPPERCASE       0x00000010 /* Force all names and attributes into upper-case characters */
#define XMF_INDENT          0x00000020 /* Indent the XML tags as they appear in the hierarchy when building strings */
#define XMF_TRANSLATE       0x00000040 /* Translate object references when building an XML string */

#define XMF_INCLUDESIBLINGS 0x80000000 /* Include siblings when building an XML string (GetXMLString only) */

/*****************************************************************************
** Methods
*/

#define MT_SetXMLAttribute -1
#define MT_GetXMLString    -2
#define MT_AddXMLStatement -3
#define MT_GetXMLContent   -4
#define MT_SortXML         -5
#define MT_RemoveXMLTag    -6

struct mtAddXMLStatement {
   LONG   Index;          /* The tag index at which the statement should be inserted (zero is the beginning) */
   STRING Statement;      /* The statement to process */
   struct XMLTag *Result; /* The resulting tag pointer */
};

struct mtGetXMLContent {
   LONG   Index;          /* Index of the tag to retrieve */
   STRING Buffer;         /* String buffer to contain the results */
   LONG   Length;         /* The size of the string buffer */
};

struct mtGetXMLString {
   LONG   Index;
   LONG   Flags;
   STRING Result;
};

struct mtRemoveXMLTag {
   LONG Index;           /* Index of the tag to remove */
   LONG Total;           /* Total number of neighbouring tags to remove */
};

struct mtSetXMLAttribute {
   LONG   Index;
   LONG   Attrib;
   STRING Name;
   STRING Value;
};

struct mtSortXML {
   LONG Index;            /* Sort everything under the specified tag, or NULL for the top level */
   struct XMLSort *List;  /* Pointer to an array of sorting information */
};

struct XMLSort {
   STRING TagName;     /* The name of the tag containing the sort information ("Default") if all sorting is based on top level tags */
   STRING Attribute;   /* The attribute to sort on, or NULL to sort on tag content */
   LONG   Flags;       /* Refer to XSF flags */
};

#define XSF_DESC         0x00000001 /* Descending */
#define XSF_CHECKSORT    0x00000002 /* Check for a 'sort' attribute in each tag and use that instead of the default sorting attribute, where possible */
#define XSF_REPORTSORTED 0x00000004
#define XSF_DESCENDING XSF_DESC

#endif /* DATA_XML_H */
