#ifndef TOOLS_DIALOG_H
#define TOOLS_DIALOG_H 1

/*
**  dialog.h
**
**  (C) Copyright 2003-2004 Rocklyte Systems.
**      All Rights Reserved.
*/

#ifndef MAIN_H
#include <pandora/main.h>
#endif

/*****************************************************************************
** Dialog class
*/

struct Dialog {
   OBJECT_HEADER             /* Class header */
   OBJECTID WindowID;        /* Reference to the window of the dialog box */
   OBJECTID TargetID;        /* The target drawable for the window (usually the desktop) */
   LONG     Flags;           /* Special options */
   LONG     Response;        /* The resulting response from the user */
   OBJECTID ResponseObject;  /* Object to update with the response result */
   LONG     Static;          /* TRUE if object is static */

   PRIVATE_FIELDS
   UBYTE    Title[80];       /* Window title */
   UBYTE    Image[100];      /* Image location */
   UBYTE    Icon[100];       /* Window icon location */
   UBYTE    InputText[120];      /* Input area text string */
   UBYTE    InputResponse[120];  /* Input area text string (response from user) */
   FIELD    ResponseField;   /* The name of the field to write the response result */
   STRING   String;          /* String to display in the dialog window */
   LONG     ButtonOffset;    /* Current horizontal offset of the buttons */
   OBJECTID TextID;          /* Reference to the text object in the dialog window */
   OBJECTID ImageID;         /* Reference to the image object in the dialog window */
   OBJECTID InputID;         /* Reference to the input box's text object, for when input boxes are used */
   struct   {
      LONG Response;         /* Response to return from the button.  May be NULL for 'do nothing' */
      UBYTE Text[40];        /* Text to print inside the button.  Button is ignored if there is no text */
   } Buttons[6];
   WORD     Active;
   WORD     TotalButtons;
   STRING   Fields[10];      /* Unlisted fields are stored here */
};

/*****************************************************************************
** Response flags
*/

#define RSF_CANCEL  0x00000001
#define RSF_YES     0x00000002
#define RSF_NO      0x00000004
#define RSF_OKAY    0x00000008
#define RSF_QUIT    0x00000010
#define RSF_NOALL   0x00000020
#define RSF_YESALL  0x00000040
#define RSF_NONE    0x00000080 /* 'Do nothing' dummy flag */

#define RSF_POSITIVE (RSF_YES|RSF_YESALL|RSF_OKAY)
#define RSF_NEGATIVE (RSF_CANCEL|RSF_NO|RSF_QUIT|RSF_NOALL)

/*****************************************************************************
** Dialog flags.
*/

#define DF_WAIT          0x00000001 /* Halt the process and wait for a user response */
#define DF_INPUT         0x00000002 /* Create an input box in the dialog window */
#define DF_INPUTREQUIRED 0x00000004 /* User input is compulsory */
#define DF_TRANSLATE     0x00000008 /* Translate the dialog String prior to display (for static dialogs only) */

#endif /* TOOLS_DIALOG_H */
