
/*****************************************************************************
** Internal: PlayPrevious()
*/

static void PlayPrevious(void)
{
   DPrintF(ProgName, "Playing the previous file in the play list.");

   if (glTrack > 0) {
      PlayTrack(glTrack-1);
   }
   else {
      /*** Restart the current track from the beginning ***/

      if (glState IS STATE_PLAYING) PlayAudio();
      else StopPlaying();
   }
}

/*****************************************************************************
** Internal: SeekAudio()
*/

static void SeekAudio(DOUBLE Position)
{
   ActionTags(AC_Seek, glCurrent, Position, POS_START);
}

/*****************************************************************************
** Internal: PlayAudio()
*/

static void PlayAudio(void)
{
   DPrintF("~PlayAudio()","Playing audio.");

   if (glCurrent) {
      if (glState IS STATE_PAUSED) {
         if (Action(AC_Enable, glCurrent, NULL) IS ERR_Okay) {
            glState = STATE_PLAYING;
         }
      }
      else {
         SeekAudio(0);
         if (Action(AC_Activate, glCurrent, NULL) IS ERR_Okay) {
            glState = STATE_PLAYING;
         }
      }

      if (glState IS STATE_PLAYING) {
         UBYTE buffer[200];
         StrFormat(buffer, sizeof(buffer), "Playing: %s", glTitle);
         SetString(buffer);
         ActionMsg(AC_Draw, glOutputID, NULL);
      }
   }
   else DPrintF("@PlayAudio:","No audio has been loaded into the player for playback.");

   StepBack();
}

/*****************************************************************************
** Internal: Pause()
*/

static void Pause(void)
{
   DPrintF("~Player:","Pausing audio playback.");

   if (glCurrent) {
      if (Action(AC_Disable, glCurrent, NULL) IS ERR_Okay) {
         SetString("Paused...");
         ActionMsg(AC_Draw, glOutputID, NULL);
         glState = STATE_PAUSED;
      }
   }

   StepBack();
}

/*****************************************************************************
** Internal: StopPlaying()
*/

static void StopPlaying(void)
{
   DPrintF("~Player:","Stopping audio playback.");

   if (glCurrent) {
      if (Action(AC_Deactivate, glCurrent, NULL) IS ERR_Okay) {
         SetString("Player Stopped");
         ActionMsg(AC_Draw, glOutputID, NULL);
         glState = STATE_STOPPED;
      }
   }

   StepBack();
}

/*****************************************************************************
** Internal: OpenFile()
*/

static ERROR OpenFile(STRING File)
{
   CLASSID classid;
   STRING title;
   WORD i;

   DPrintF(ProgName, "Opening file \"%s\"", File);

   if (glCurrent) {
      if (Action(AC_Deactivate, glCurrent, NULL) IS ERR_Okay) {
         SetString("Player Stopped");
         glState = STATE_STOPPED;
      }

      acFree(glCurrent);
      glCurrent = NULL;
   }

   if (IdentifyFile(File, &classid, NULL) IS ERR_Okay) {
      if (classid IS ID_MUSIC) {
         if (CreateObject(ID_MUSIC, NULL, &glCurrent, NULL,
            FID_Location|TSTRING, File,
            TAGEND) IS ERR_Okay) {
         }
      }
      else if (classid IS ID_SOUND) {
         if (CreateObject(ID_SOUND, NULL, &glCurrent, NULL,
            FID_Location|TSTRING, File,
            TAGEND) IS ERR_Okay) {
         }
      }
   }

   if (glCurrent) {
      if ((GetField(glCurrent, FID_Title, FT_STRING, &title) IS ERR_Okay) AND (title)) {
         StrCopy(title, glTitle, sizeof(glTitle));
      }
      else {
         for (i=0; File[i]; i++);
         while ((i > 0) AND (File[i-1] != ':') AND (File[i-1] != '/') AND (File[i-1] != '\\')) i--;
         StrCopy(File + i, glTitle, sizeof(glTitle));
      }
   }
   else {
      OBJECTPTR dialog;

      if (CreateObject(ID_DIALOG, NULL, &dialog, NULL,
         FID_Buttons|TSTRING, "okay",
         FID_Title|TSTRING,   "File Not Recognised",
         FID_Image|TSTRING,   "icons:items/error(48)",
         FID_String|TSTRING,  "The file is not in a recognisable audio format.  You may need to install an audio plugin to play this file type.",
         TAGEND) IS ERR_Okay) {
         Action(AC_Show, dialog, NULL);
      }
   }

   ActionMsg(AC_Draw, glOutputID, NULL);

   if (glCurrent) return(ERR_Okay);
   else return(ERR_Failed);
}

/*****************************************************************************
** Internal: GetTitle()
*/

static STRING GetTitle(OBJECTPTR Object, STRING Filename)
{
   STRING title;
   WORD i;

   if ((Object) AND (GetField(Object, FID_Title, FT_STRING, &title) IS ERR_Okay) AND (title) AND (title[0])) {
      return(StrClone(title));
   }
   else {
      for (i=0; Filename[i]; i++);
      while ((i > 0) AND (Filename[i-1] != ':') AND (Filename[i-1] != '/') AND (Filename[i-1] != '\\')) i--;
      return(StrClone(Filename + i));
   }
}

/*****************************************************************************
** Internal: AddViewXML()
*/

static void AddViewXML(WORD Index)
{
   UBYTE buffer[1024];

   StrFormat(buffer, sizeof(buffer), "<item file=\"%s\" track=\"%d\">%s<track type=\"numeric\">%d</track><length>%d:%.2d</length></item>",
      glPlaylist[Index].File, Index + 1, glPlaylist[Index].Title, Index + 1, glPlaylist[Index].Length / 60, glPlaylist[Index].Length % 60);

   DPrintF("AddViewXML()","%60s", buffer);
   ActionTags(AC_DataChannel, glPlaylistView, glTask->UniqueID, DATA_XML, 1, buffer, StrLength(buffer), 1);
}

/*****************************************************************************
** Internal: PlayTrack()
** Short:    Loads the track and starts playing it.
*/

static void PlayTrack(WORD Track)
{
   DPrintF("PlayTrack:","Track: %d", Track);

   if ((Track < 0) OR (Track > ARRAYSIZE(glPlaylist))) return;

   if (OpenFile(glPlaylist[Track].File) IS ERR_Okay) {
      PlayAudio();
      glTrack = Track;
   }
   else StopPlaying();
}

/*****************************************************************************
** Internal: LoadTrack()
** Short:    Loads the track but doesn't play it.
*/

static void LoadTrack(WORD Track)
{
   DPrintF("LoadTrack:","Track: %d", Track);

   if ((Track < 0) OR (Track > ARRAYSIZE(glPlaylist))) return;

   if (OpenFile(glPlaylist[Track].File) IS ERR_Okay) {
      glTrack = Track;
   }
}

/*****************************************************************************
** Internal: GetXMLAttribute()
*/

static STRING GetXMLAttribute(struct XMLTag *Tag, STRING Attribute)
{
   WORD i;

   for (i=0; i < Tag->TotalAttributes; i++) {
      if (StrCompare(Tag->Attributes[i].Name, Attribute, 0, STR_MATCHLENGTH) IS ERR_Okay)  {
         return(Tag->Attributes[i].Value);
      }
   }
   return(NULL);
}

/*****************************************************************************
** Internal: GetXMLAttributeInt()
*/

static LONG GetXMLAttributeInt(struct XMLTag *Tag, STRING Attribute)
{
   WORD i;

   for (i=0; i < Tag->TotalAttributes; i++) {
      if (StrCompare(Tag->Attributes[i].Name, Attribute, 0, STR_MATCHLENGTH) IS ERR_Okay)  {
         return(StrToInt(Tag->Attributes[i].Value));
      }
   }
   return(0);
}
